/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */
 
/*
 * CC1000 Frequency Hopping Library
 *
 * This library allows for fast channel switching of pre-configured channels
 *
 * Usage:
 * - cc1000_freqhop_init defines the used channels
 * - cc1000_freqhop_rxmode and cc1000_freqhop_txmode are used to switch to the specified
 *         and sets the radio mode
 *
 * Author: Matthias Ringwald <mringwal@inf.ethz.ch>
 *
 */

#ifdef __BTNODE3__

#include <sys/types.h>
#include <cc/cc1000.h>
#include <cc/cc1000_defs.h>
#include <cc/cc1000_params.h>
#include <cc/cc1000_freqhop.h>

#include <stdio.h>

/** number of channels */
static u_char numChannels;

/** current channel */
static u_char currentChannel = 255;

/** calibrated channel table */
static struct cc_channel * channels;

/** private access to cc1000.c:cc1000_calibrate() results */ 
extern u_char vco_array_rx;
extern u_char vco_array_tx;
extern u_char chp_current_rx;
extern u_char chp_current_tx;

#if 0
/**
 * Dump calculated and calibrated channel data
 */
void printChannel(struct channel *currChannel){
    printf("%9lu: %9lu %02x:%02x %02x %02x:%02x:%02x %02x:%02x:%02x %02u %02u %02u %02u\n",
           currChannel->base,
           currChannel->actFreq,
           currChannel->fsep[0],   
           currChannel->fsep[1],
           currChannel->refdiv,   
           currChannel->freqA[0],
           currChannel->freqA[1],
           currChannel->freqA[2],
           currChannel->freqB[0],
           currChannel->freqB[1],
           currChannel->freqB[2],
           currChannel->chp_co_rx, 
           currChannel->vco_ao_rx,
           currChannel->chp_co_tx, 
           currChannel->vco_ao_tx);
}
#endif

/**
 * Init frequency table and calibrate each RX/TX pair
 *
 * @param frequency number of channels to be used
 * @param frequency of channel #0
 * @param frequency seperation between two adjunct channels
 *
 * @note  It is assumed that this function is only called once
 *        therefore it does not bother with memory freeing
 *
 * @pre   cc1000 is initialized
 * @post  calibrated frequency table is ready
 */
void cc1000_freqhop_init(u_char nr, u_long base, u_long seperation) {
    
    int i;
    struct cc_channel * currChannel;
    
    // alloc memory
    channels = malloc ( nr * sizeof(struct cc_channel));
    numChannels = nr;

    // set up channels
    for (i = 0; i < nr; i++) {
        currChannel = &channels[i];

        // calculate channel configuration
        cc1000_calc_channel( base, currChannel);

        // tune to this channel
        cc1000_set_channel(currChannel);
        
        // calibrate
        cc1000_calibrate();

        // store calibration results from CC1000_calibrate();
        currChannel->chp_co_rx = chp_current_rx | (1 << CC1000_CHP_OVERRIDE);
        currChannel->chp_co_tx = chp_current_tx | (1 << CC1000_CHP_OVERRIDE);
        currChannel->vco_ao_rx = vco_array_rx   | (1 << CC1000_VCO_OVERRIDE);
        currChannel->vco_ao_tx = vco_array_tx   | (1 << CC1000_VCO_OVERRIDE);

        // next frequency
        base += seperation;
    }
}


/**
 * Set channel for receive
 *
 * @param channel nr
 * @pre   cc1000_freqhop_init(channels, ..) was called and nr < channels
 * @pre   basic configuration was already done
 * @post  cc1000 registers are set. mode is not set
 */
void cc1000_freqhop_rxmode(u_char nr) {
    struct cc_channel *currChannel = &channels[nr];
    u_char current = cc1000_params_template[RXCURRENT_INDEX];

    // set up registers for channel
    if (nr != currentChannel) {
        cc1000_set_channel( currChannel );
        currentChannel = nr;
    }
    
    // Procedures on the data sheet 
    // Turn on RX: RX_PD = 0, FS_PD = 0 RX_PD = 0, FS_PD = 0
    cc1000_write(CC1000_MAIN, ((1 << CC1000_TX_PD) | (1 << CC1000_RESET_N)));

    // Writing current
    cc1000_write(CC1000_CURRENT, current);

    // write calibrated data
    cc1000_write(CC1000_TEST6, currChannel->chp_co_rx);
    cc1000_write(CC1000_TEST5, currChannel->vco_ao_rx);

    // Wait 250 us
    delay_time(250);

}

/**
 * Set channel for transmit
 *
 * @param channel nr
 * @pre   cc1000_freqhop_init(channels, ..) was called and nr < channels
 * @pre   basic configuration was already done
 * @post  cc1000 registers are set. mode is not set
 */
void cc1000_freqhop_txmode(u_char nr) {
    u_char rfPower;
    struct cc_channel *currChannel = &channels[nr];
    u_char current = cc1000_params_template[TXCURRENT_INDEX];

    // get rf power set in cc1000.c
    cc1000_get_RF_power(&rfPower);

    // set up registers for channel
    if (nr != currentChannel) {
        cc1000_set_channel( currChannel );
        currentChannel = nr;
    }
    
    // Procedures on the data sheet 
    // Turn on TX: PA_POW = 00h MAIN: RXTX = 1, F_REG = 1 TX_PD = 0, FS_PD = 0
    cc1000_write(CC1000_MAIN, ((1 << CC1000_RXTX) | (1 << CC1000_F_REG) | (1 << CC1000_RX_PD) | (1 << CC1000_RESET_N)));

    // Writing current
    cc1000_write(CC1000_CURRENT, current);

    // write calibrated data
    cc1000_write(CC1000_TEST6, currChannel->chp_co_tx);
    cc1000_write(CC1000_TEST5, currChannel->vco_ao_tx);

    // Wait 250 us
    delay_time(250);

    // PA_POW =  Output power 
    cc1000_write(CC1000_PA_POW, rfPower);

    // Wait 250 us
    delay_time(250);
}


/** 
 * Get access to calibrated channel struct *
 *
 * @param channel nr
 * @return pointer to requested channel struct 
 */
struct cc_channel * getChannel(u_char nr){
	return  &channels[nr];
}

#endif
