/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */
#ifndef RPC_H_
#define RPC_H_

#include <bt/bt_psm.h>
#include <mhop/mhop_cl.h>


#define RPC_ERR_TOO_MANY_PROCS              -1

////#define RPC_MAX_RESULT_LENGTH (MHOP_CL_PKT_MAX_PAYLOAD - 1)
//#define RPC_MAX_RESULT_LENGTH 80

/** 
 * \brief definition of a rpc procedure that can be registered. Such a procedure
 * reads and interprets the query and write its results in the provided buffer.
 * 
 * If the result is longer than rpc_max_result_size(), the function should split
 * up the result in junks and should return 1. The function will be called 
 * again such that the next junk can be stored in the result buffer, until
 * 0 is returned.
 * 
 * \param[in] query the query data
 * \param[in] query_len length of query
 * \param[in] arg callback argument that has been given to the registration
 * \param[in] result pointer to the result buffer, where the function can store
 *    the results
 * \param[out] result_len length of the stored result, max. 
 * rpc_max_result_size()
 * \return 1 if more than one pkt has to bee sent, else 0.
 * \see rpc_register_proc(), rpc_max_result_size()
 */
#define RPC_PROC \
    u_char (*rpc_proc) (bt_acl_pkt_buf* pkt_buf, \
    					char* query, \
    					u_char query_len, \
    					void* arg, \
                        u_char* result, \
                        u_char* result_len) \


#define RPC_RESULT \
    void (*rpc_result) (u_char* source_addr, u_char* result, \
                        u_char result_len, void* arg)

typedef struct _rpc_procs_s {
    RPC_PROC;
    void * proc_arg;
    RPC_RESULT;
    void * result_arg;
} _rpc_procs_t;

typedef struct _rpc_stack_s {
    _rpc_procs_t* procs; 
    u_char max_procs;
    u_short proc_psm;
    u_short result_psm;
    u_char* result_data;
} rpc_stack_t;
    
rpc_stack_t* _rpc;

/**
 * \brief Registers a RPC procedure and a result callback.
 * \param proc_nr procedure nummer assigned by this function
 * \param RPC_PROC procedure callback
 * \param proc_arg procedure argument (will be passed to the callback)
 * \param RPC_RESULT result callback
 * \param result_arg result argument (will be passed to the callback)
 * \return 0 on success, or RPC_ERR_TOO_MANY_PROCS
 */    
char rpc_register_proc(u_char* proc_nr, \
                       RPC_PROC, void* proc_arg, \
                       RPC_RESULT, void* result_arg);

/**
 * \brief sends a rpc query.
 * \param proc_nr rpc nummer of the procedure to be called
 * \param data query data that will be sent to the rpc proc
 * \param data_len length of data, maximal rpc_max_result_size()
 * \param dest_addr destination address or bt_addr_null for broadcast
 * \param ttl time to life in hops, 0 for infinity
 */
void rpc_send_query(u_char proc_nr, u_char* data, u_char data_len,
                    bt_addr_t dest_addr, u_char ttl);

/**
 * \brief calls a registered rpc procedure locally.
 * \param proc_nr rpc nummer of the procedure to be called
 * \param data query data that will be sent to the rpc proc
 * \param data_len length of data, maximal rpc_max_result_size()
 */
void rpc_local_query(u_char proc_nr, u_char* data, u_char data_len);

/**
 * \brief returns the maximal size for query and result data.
 */
inline u_short rpc_max_result_size(void);    

/**
 * \brief initializes the rpc stack.
 * \param psmux pointer to the psmux stack
 * \param nr_procs maximal nummer of supported procedures
 * \param proc_psm PSM of the rpc procedure callback
 * \param result_psm PSM of the rpc result callback
 */                      
rpc_stack_t* rpc_init(bt_psm_t* psmux,
						u_char nr_procs,
						u_short proc_psm,
						u_short result_psm);

#endif /*RPC_H_*/
