#ifndef _VC_H_
#define _VC_H_

/**
 * \file vcp.h
 * \brief Virtual Connection Protocol Layer
 */
#include <sys/types.h>
#include <bt/bt_defs.h>
#include <mhop/tp.h>
#include <mhop/ppp.h>


// service number
#define VCP_PPP_PORT 3

#define VCP_NR_PORTS 10
/**
 * \name VC Packets
 */
 //@{ 
/// virtual connection data packet header length
#define VCP_HEADER_LEN (2+2)					// +1 for ppp port
// max payload
#define VCP_MAX_PAYLOAD PPP_MAX_PAYLOAD - VCP_HEADER_LEN
//@}

/**
 * \name TP error codes
 * Returned by tp_send_*_data() functions on failure.
 */
//@{ 
/// Send error
#define VCP_ERR_SEND     1
/// Packet size too large
#define VCP_ERR_PKT_LEN  2
/// Destination invalid
#define VCP_ERR_DEST     3
//@}

/** \name VC connection callback types and reasons */
//@{ 
/// VC type: open
#define VCP_CON_OPEN        1
/// VC type: close
#define VCP_CON_CLOSE       2
/// VC close reason: link failure
#define VCP_CLOSE_NORMAL 1 
/// VC close reason: normal close (by other side) 
#define VCP_CLOSE_LINK   2 
//@}

////////////////////////////////////////////////////////////////////////////////
/// \name Transport initialization functions
////////////////////////////////////////////////////////////////////////////////
//@{

/**
 * \brief Initializes the transport layer.
 * \param bt_stack Pointer to the btstack.
 * \param acl_com_stack Pointer to the acl com stack.
 * \param rel_con_change_registration Registration function for reliable 
 *        connection changes.
 */
void tp_init(
        struct btstack *bt_stack,
        void (*rel_con_change_registration)(void (*cb)(u_char type, u_char detail, bt_con_handle_t con_handle))
);

/**
 * \brief Registration function for vc data callback. 
 * The registered function is called on arrival of vc data.
 */
void tp_register_vc_data_cb(void (*cb)(u_char vc_handle, u_char* data, u_short len));

/**
 * \brief Registration function for connection less data callback. 
 * The registered funciton is called on arrival of cl data.
 * 
 * \param cb Callback function to register. Fallowing arguments are passed to
 *        this function:
 *        \arg \c bt_addr_t \e source Address of the source node
 *        \arg \c u_char* \e data Pointer to the begin of data
 *        \arg \c u_short \e len Length of data in bytes
 */
//void tp_register_cl_data_cb(void (*cb)(bt_addr_t source, u_char* data, u_short len));

#define VCP_DATA_CB void (*data_cb) (u_char vc_handle, u_char* data, u_short len, u_char port_nr);
/**
 * \brief Registration function for vc connection changes. 
 * The registered function is called if a new vc opens or an existing vc was 
 * closed.
 */
#define VCP_VC_CHANGE_CB void (*vc_change_cb) (u_char vc_handle, u_char type, u_char reason);

// to register an upper layer service
short vcp_register_service(u_char port, VCP_DATA_CB, VCP_VC_CHANGE_CB, u_char pkt_processors);
//@}

////////////////////////////////////////////////////////////////////////////////
/// \name VCP API functions
////////////////////////////////////////////////////////////////////////////////
//@{

/**
 * \brief Opens a virtual connection to the target with address \a target.
 * \param target Address of the target
 * \param ttl Time to live
 */
void vcp_open_vc(bt_addr_t target, u_char ttl);

/**
 * \brief Closes the virtual connection with handle \a vc_handle.
 * \param vc_handle VC handle of the virtual connection to close
 */
void vcp_close_vc(u_char vc_handle);

/**
 * \brief Sends a VC data packet over the virtual connection with handle 
 * \a vc_handle.
 * 
 * To send a VC data packet, first use tp_get_vc_data_pointer(), fill in the data
 * and then use this function.
 * 
 * \param pkt Pointer to the bt_l2cap_pkt_t packet to send
 * \param vc_handle VC handle of the virtual connection to use
 * \param len Length of the data to send in bytes
 * \return 0 on success, on error: TP_ERR_DEST, TP_ERR_PKT_LEN, TP_ERR_SEND
 */
u_char vcp_send_data(vcp_pkt_t* pkt, u_char vc_handle, u_short len);

/**
 * \brief Returns a pointer to the begin of payload data in a VC data packet
 * 
 * \param pkt Pointer to the bt_l2cap_pkt_t packet
 * \return Pointer to the begin of the VC payload data in \p pkt 
 */
u_char* vcp_get_data_pointer(vcp_pkt_t* pkt);
//@}
#endif //_VC_H_
